from langchain_community.tools.pubmed.tool import PubmedQueryRun
from demogpt_agenthub.tools import BaseTool

class PubmedTool(BaseTool):
    def __init__(self):
        self.tool = PubmedQueryRun()
        super().__init__()
        
    def run(self, query):
        return self.tool.invoke(query)

if __name__ == "__main__":
    tool = PubmedTool()
    print(tool.run("covid"))


from demogpt_agenthub.agents import BaseAgent

class ReactAgent(BaseAgent):
    def __init__(self, tools, llm, verbose=False, max_iter=10):
        super().__init__(tools, llm, verbose, max_iter)
    
    def run(self, prompt):
        self.add_message("User", prompt)
        result_ready = False
        iter = 0
        while not result_ready and iter < self.max_iter:
            result_ready = self.success_decider.invoke({"task": prompt, "context": self.context, "tools": self.tool_explanations})
            self.pretty_print("Decision", result_ready)
            if result_ready:
                break
            decision = self.tool_decider.invoke({"task": prompt, "context": self.context, "tools": self.tool_explanations})
            self.add_message("Agent", decision["reasoning"])
            self.pretty_print("Reasoning", decision["reasoning"])
            self.pretty_print("Tool call", decision["tool"])
            tool_call = self.tools[decision["tool"]]
            tool_args = decision["argument"]
            self.pretty_print("Tool args", tool_args)
            tool_result = tool_call.run(**tool_args)
            self.add_message(decision["tool"], tool_result)
            self.pretty_print("Tool result", tool_result)
            iter += 1
        if not result_ready:
            self.pretty_print("Not Completed", """The task was not completed within the maximum number of iterations. The agent will try to answer with the available context.
                              If you want to try again, you can increase the maximum number of iterations by setting the max_iter parameter when creating the agent.""")
        answer = self.final_answer.invoke({"query": prompt, "context": self.context})
        self.add_message("Agent", answer)
        self.pretty_print("Answer", answer)
        return answer

if __name__ == "__main__":
    from demogpt_agenthub.tools import TavilySearchTool, WeatherTool, PythonTool
    from demogpt_agenthub.llms import OpenAIChatModel
    search_tool = TavilySearchTool()
    weather_tool = WeatherTool()
    python_tool = PythonTool()
    agent = ReactAgent(tools=[search_tool, weather_tool, python_tool], llm=OpenAIChatModel(model_name="gpt-4o-mini"), verbose=True)
    query = "What is the weather's temperature's square root in the country where Christiano Ronaldo is currently playing? Please precisely calculate the result."
    print(agent.run(query))


from langchain_community.utilities import StackExchangeAPIWrapper
from demogpt_agenthub.tools import BaseTool

class StackOverFlowTool(BaseTool):
    def __init__(self):
        self.tool = StackExchangeAPIWrapper()
        super().__init__()
        
    def run(self, query):
        return self.tool.run(query)
    
if __name__ == "__main__":
    query = "How to create a list in python"
    tool = StackOverFlowTool()
    print(tool.run(query))


system_template = """
You are a helpful assistant that's good at accomplishing tasks.
The task comes from the user and you have to decide which tool to use to accomplish the task.
You'll be given the current context and the tools that you have.
The tools consist of a name, description, and a function that can be used to call the tool.

Your response should be a valid JSON object in the following format:

{{
    "reasoning": <reasoning>,
    "tool": <tool_name>,
    "argument": <tool_argument as a dictionary of parameters>
}}
"""

human_template = """
Task: {task}
==============
Context: {context}
==============
Tools: {tools}
==============
Response JSON:
"""


from langchain_core.prompts import ChatPromptTemplate
from langchain_core.output_parsers import JsonOutputParser, StrOutputParser
from demogpt_agenthub.utils.parsers import BooleanOutputParser
from demogpt_agenthub.prompts.agents.tool_calling import tool_decider, final_answer
from demogpt_agenthub.prompts.agents.react import success_decider

from dotenv import load_dotenv
import inspect

load_dotenv()

class BaseAgent:
    def __init__(self, tools, llm, verbose=False, max_iter=10):
        self.llm = llm
        self.tool_decider_prompt = ChatPromptTemplate.from_messages([
            ("system", tool_decider.system_template),
            ("human", tool_decider.human_template)
        ])
        self.tool_decider = self.tool_decider_prompt | self.llm | JsonOutputParser()
        self.final_answer_prompt = ChatPromptTemplate.from_messages([
            ("system", final_answer.system_template),
            ("human", final_answer.human_template)
        ])
        self.final_answer = self.final_answer_prompt | self.llm | StrOutputParser()

        self.success_decider_prompt = ChatPromptTemplate.from_messages([
            ("system", success_decider.system_template),
            ("human", success_decider.human_template)
        ])

        self.success_decider = self.success_decider_prompt | self.llm | BooleanOutputParser()

        self.history = []
        self.tools = tools
        self.verbose = verbose
        self.tools = {tool.name: tool for tool in tools}
        self.max_iter = max_iter

    @property
    def tool_explanations(self):
        explanations = {}
        for tool in self.tools.values():
            # Get the run function signature
            sig = inspect.signature(tool.run)
            # Format parameters, excluding 'self'
            params = [f"{name}: {param.annotation.__name__ if param.annotation != inspect._empty else 'any'}"
                     for name, param in sig.parameters.items() if name != 'self']
            param_str = ", ".join(params)
            
            explanations[tool.name] = {
                "description": tool.description,
                "run_description": tool.run.__doc__ or "No description available",
                "parameters": f"Input parameters: ({param_str})"
            }
        return explanations

    @property
    def context(self):
        return "\n".join([f"{name}: {message}" for name, message in self.history])
    
    def add_message(self, name, message):
        self.history.append((name, message))

    def pretty_print(self, message_type, content):
        """
        Displays the given content with color-coded formatting based on the message type.

        Args:
            message_type (str): The type of message (e.g., "Decision", "Tool call", "Tool result", "Answer").
            content (str): The content to be displayed.
        """

        if not self.verbose:
            return

        color_codes = {
            "Decision": "\033[95m",  # Magenta
            "Reasoning": "\033[94m",  # Blue
            "Tool call": "\033[93m",  # Yellow  
            "Tool args": "\033[93m",  # Yellow
            "Tool result": "\033[92m",  # Green
            "Not Completed": "\033[91m",  # Red
            "Answer": "\033[92m"  # Green
        }

        if message_type in color_codes:
            print(f"{message_type}:")
            print(f"{color_codes[message_type]}{content}\033[0m")
        else:
            print(f"{message_type}:")
            print(content)
    
    def run(self, prompt: str):
        """
        Abstract method to process a user's prompt and return a response.

        This method should be implemented by subclasses to define the specific
        behavior of how the agent processes and responds to user prompts.

        Args:
            prompt (str): The user's input prompt or question.

        Returns:
            str: The agent's response to the user's prompt.

        Raises:
            NotImplementedError: If the subclass does not implement this method.
        """
        raise NotImplementedError("Subclasses must implement the 'ask' method.")


system_template = """
You are a helpful assistant that can answer the user query based on the provided context.
The context is a list of messages between the user and the assistant.
The assistant tried to find the answer using various tools.
Your job is to answer the user query based on the context.
Asnwer as a kind assistant.
"""

human_template = """
User: {query}
=============
Context:
{context}
=============
Answer:
"""



from langchain_community.tools import YouTubeSearchTool as Youtube
from demogpt_agenthub.tools import BaseTool

class YouTubeSearchTool(BaseTool):
    def __init__(self):
        self.tool = Youtube()
        super().__init__()

    def run(self, query):
        return self.tool.run(query)
    
if __name__ == "__main__":
    query = "lex friedman"
    tool = YouTubeSearchTool()
    print(tool.run(query))


class BaseTool:
    def __init__(self, **data):
        super().__init__(**data)
        self.setAttributes()

    def setAttributes(self):
        if hasattr(self, 'tool') and self.tool:
            for attribute in self.tool.__dict__:
                setattr(self, attribute, self.tool.__dict__[attribute])


from langchain_openai import OpenAI, ChatOpenAI
from demogpt_agenthub.llms.base import BaseLLM

class OpenAIModel(OpenAI, BaseLLM):
    def run(self, prompt: str) -> str:
        return self.invoke(prompt)

class OpenAIChatModel(ChatOpenAI, BaseLLM):
    def run(self, prompt: str) -> str:
        return self.invoke(prompt).content


from langchain_experimental.utilities import PythonREPL
from demogpt_agenthub.tools import BaseTool

class PythonTool(BaseTool):
    def __init__(self):
        self.tool = PythonREPL()
        super().__init__()
        self.name = "Python Interpreter"
        self.description = """A tool that can execute Python code to perform precise calculations by using the Python programming language. For any calculations that require a high degree of precision, this tool must be used.
        You must add print statement to the code to see the results of the calculations. Otherwise, the result will not be displayed."""
        
    def run(self, code):
        return self.tool.run(code)
    
if __name__ == "__main__":
    tool = PythonTool()
    code = "print('Hello, World!')"
    print(tool.run(code))


from demogpt_agenthub.tools.base import BaseTool
from demogpt_agenthub.tools.tavily_search import TavilySearchTool
from demogpt_agenthub.tools.bash import BashTool
from demogpt_agenthub.tools.repl import PythonTool
from demogpt_agenthub.tools.research import ArxivTool
from demogpt_agenthub.tools.youtube import YouTubeSearchTool
from demogpt_agenthub.tools.stack_exchange import StackOverFlowTool
from demogpt_agenthub.tools.req import RequestUrlTool
from demogpt_agenthub.tools.wikiped import WikipediaTool
from demogpt_agenthub.tools.wikidata import WikiDataTool
from demogpt_agenthub.tools.pubmed import PubmedTool
from demogpt_agenthub.tools.weather import WeatherTool # requires OPENWEATHERMAP_API_KEY
from demogpt_agenthub.tools.yolo import YoloTool